"""Train brain wave decoder."""
import os
import sys
from typing import Any, Dict, Mapping, Optional, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import optax
from flax.core.frozen_dict import FrozenDict
from tqdm import tqdm

sys.path.append(".")

from src.load_eeg import load_train_valid_test


def normalize(
    data: np.ndarray, mean: Optional[float] = None, std: Optional[float] = None
) -> Tuple[np.ndarray, float, float]:
    """Normalize the input array.

    After normalization the input
    distribution should be approximately standard normal.

    Args:
        data (np.array): The input array.
        mean (float): Data mean, re-computed if None.
            Defaults to None.
        std (float): Data standard deviation,
            re-computed if None. Defaults to None.

    Returns:
        np.array, float, float: Normalized data, mean and std.
    """
    if mean is None:
        mean = np.mean(data)
    if std is None:
        std = np.std(data)
    return (data - mean) / std, mean, std


class BrainCNN(nn.Module):
    """Your Brain-CNN model."""

    @nn.compact
    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        """Run the forward pass."""
        x = nn.Conv(features=128, kernel_size=(1, 3))(x)
        x = nn.relu(x)
        x = nn.Conv(features=256, kernel_size=(1, 3))(x)
        x = nn.relu(x)
        x = nn.max_pool(x, window_shape=(1, 3))
        x = x.reshape((x.shape[0], -1))  # flatten
        x = nn.Dense(features=4)(x)
        return x


@jax.jit
def forward_step(
    variables: FrozenDict, img_batch: jnp.ndarray, label_batch: jnp.ndarray
):
    """Do a forward step."""
    out = cnn.apply(variables, img_batch)
    ce_loss = jnp.mean(
        optax.softmax_cross_entropy(
            logits=out, labels=nn.one_hot(label_batch, num_classes=4)
        )
    )
    return ce_loss


@jax.jit
def adjust_dimensions(network_input: jnp.ndarray) -> jnp.ndarray:
    """Adjust the eeg_input dimensions to allow CNN processing.

    Args:
        network_input (jnp.ndarray): The eeg input array.

    Returns:
        jnp.ndarray: An output array suitable to a CNN's liking.
    """
    return jnp.expand_dims(jnp.array(network_input).transpose(0, 2, 1), -3)


def get_acc(
    net: nn.Module,
    variables: Union[FrozenDict[str, Mapping[str, Any]], Dict[str, Any]],
    eeg_input: jnp.ndarray,
    labels: jnp.ndarray,
) -> jnp.ndarray:
    """Compute the accuracy.

    Args:
        net (nn.Module): Your cnn object.
        variables (FrozenDict): The network weights.
        eeg_input (jnp.ndarray): An array containing the eeg brain waves.
        labels (jnp.ndarray): The action annotations in a array.

    Returns:
        jnp.ndarray: The accuracy in [%].
    """
    logits = net.apply(variables, adjust_dimensions(eeg_input))
    accuracy = jnp.mean(jnp.argmax(logits, -1) == labels)
    return accuracy


if __name__ == "__main__":
    low_cut_hz = 0
    subject_id = 1
    batch_size = 50
    epochs = 20

    train_filename = os.path.join("./data", "train/{:d}.mat".format(subject_id))
    test_filename = os.path.join("./data", "test/{:d}.mat".format(subject_id))

    # Create the dataset
    train_set, valid_set_np, test_set_np = load_train_valid_test(
        train_filename=train_filename,
        test_filename=test_filename,
        low_cut_hz=low_cut_hz,
    )

    train_set_x, mean, std = normalize(train_set.X)
    valid_set_x_np, _, _ = normalize(valid_set_np.X, mean, std)
    test_set_x_np, _, _ = normalize(test_set_np.X, mean, std)

    train_size = train_set.X.shape[0]
    train_input = np.array_split(train_set_x, train_size // batch_size)
    train_labels = np.array_split(train_set.y, train_size // batch_size)

    valid_set_y = jnp.array(valid_set_np.y)
    valid_set_x = jnp.array(valid_set_x_np)
    test_set_y = jnp.array(test_set_np.y)
    test_set_x = jnp.array(test_set_x_np)

    key = jax.random.PRNGKey(42)  # type: ignore
    cnn = BrainCNN()
    opt = optax.adam(learning_rate=0.001)

    variables = cnn.init(
        key, jnp.expand_dims(jnp.ones(train_set.X.shape).transpose(0, 2, 1), -3)
    )
    opt_state = opt.init(variables)
    loss_grad_fn = jax.value_and_grad(forward_step)

    val_acc_list = []
    for e in range(epochs):
        bar = tqdm(
            zip(train_input, train_labels),
            total=len(train_input),
            desc="Training Brain CNN",
        )
        for input_x, labels_y in bar:
            input_x = adjust_dimensions(input_x)
            labels_y = jnp.array(labels_y)

            cel, grads = loss_grad_fn(variables, input_x, labels_y)
            updates, opt_state = opt.update(grads, opt_state, variables)
            variables = optax.apply_updates(variables, updates)
            bar.set_description("Loss: {:2.3f}".format(cel))

        val_accuracy = get_acc(cnn, variables, valid_set_x, valid_set_y)
        print("Validation accuracy {:2.3f} at epoch {}".format(val_accuracy, e + 1))  # type: ignore
        val_acc_list.append(val_accuracy)

    test_accuracy = get_acc(cnn, variables, test_set_x, test_set_y)
    print("Test accuracy: {:2.3f}".format(test_accuracy))  # type: ignore
    plt.plot(val_acc_list, label="Validation accuracy")
    plt.plot(len(val_acc_list) - 1, test_accuracy, ".", label="Test accuracy")
    plt.xlabel("epochs")
    plt.ylabel("accuracy")
    plt.legend()
    plt.show()
